#include "disosadapt.h"
#include "stdio.h"
#include "log.h"
#include "openssl/ssl.h"

#ifdef OS_VER
#if(OS_WIN ==OS_VER)

static HANDLE* lockarray;

unsigned int SmCreate( SM_T * const pSm )
{
	if(NULL == pSm)
    {
        return OS_ERR_INVALID_PARAM;
    }

    pSm->SmHandler = CreateMutex( NULL, FALSE, NULL );
    
    if (pSm->SmHandler != NULL)
    {
        return OS_OK;
    }
    else
    {
        return OS_ERR_NOT_ENOUGH_MEM;
    }
}

unsigned int SmDelete( SM_T *const pSm )
{
    unsigned int   err;
    if(NULL == pSm)
    {
        return OS_ERR_INVALID_PARAM;
    }

    err = (unsigned int)CloseHandle( pSm->SmHandler ); 
    if ( 0 == err )
    {
        err = OS_ERR_OBJ_ID_ERROR;
    }
    else
    {
        err = 0;
    }

    return err;
}

unsigned int SmP( SM_T*const  pSm, unsigned int ulTimeOutInMillSec )
{
    unsigned int err;

    if(NULL == pSm)
    {
        return OS_ERR_INVALID_PARAM;
    }  

    if ( 0 == ulTimeOutInMillSec )
    {
        ulTimeOutInMillSec = INFINITE;
    } 
    
    err = WaitForSingleObject( pSm->SmHandler, ulTimeOutInMillSec );
    switch( err)
    {
        case WAIT_OBJECT_0:
            err = 0;
            break;

        case WAIT_ABANDONED:
            err = OS_ERR_OBJ_UNAVAILABLE;       
            break;

        case WAIT_TIMEOUT:
            err = OS_ERR_TIMEOUT;
            break;
        default:
            err    = OS_ERR_OPFAIL;
            break;
    }

    return err;
}

unsigned int SmV( SM_T* const pSm )
{
    unsigned int err;

    if(NULL == pSm)
    {
        err = OS_ERR_INVALID_PARAM;
        return err;
    }

    /* windows系统中，该函数返回非零表示成功，0表示失败，因此为了统一描述，将修正返回值 */
    err = (unsigned int)ReleaseMutex( pSm->SmHandler );
    if( 0 == err )
    {
        err = OS_ERR_OBJ_ID_ERROR;
    }
    else
    {
        err = 0;
    }

    return err;
}



unsigned int CreateTask
(  
    void  (* pfnStartRoutine)(void*),
    void   * pArg, 
    unsigned int * pulTaskID 
)
{
    unsigned int  ulReturnCode;
    unsigned int  ulThreadId = 0;
    HANDLE  tashHandle;

    if(NULL == pulTaskID || NULL == pfnStartRoutine )
    {
        return OS_ERR_INVALID_PARAM;
    }
   
    tashHandle = CreateThread(NULL,
                              0,
                              ( LPTHREAD_START_ROUTINE )pfnStartRoutine,
                              ( void * )pArg,
                              0,
                              &ulThreadId );
    if ( NULL == tashHandle )
    {
        *pulTaskID     = 0;
        return OS_ERR_NOT_ENOUGH_MEM;
    }

    *pulTaskID = (unsigned int)tashHandle;
    return 0;
}

unsigned int DeleteTask( unsigned int ulTaskID )
{
    unsigned int ulReturnCode;
 
    ulReturnCode = (unsigned int)TerminateThread( ( HANDLE )ulTaskID, 0 );
    if ( 0 == ulReturnCode )
    {
        return OS_ERR_OPFAIL;
    }
    
    /*ExitThread((HANDLE)ulOsalTaskID);*/
    return 0;
}

static void lock_callback(int mode, int type, char *file, int line)  
{
    (void)file;
    (void)line;
    if (mode & CRYPTO_LOCK) {
        WaitForSingleObject(lockarray[type], INFINITE);//lint !e409 
    }
    else
    {
        ReleaseMutex(lockarray[type]);//lint !e409
    }  
}  

static unsigned long thread_id(void)  
{
    unsigned long ret;  
    ret = (unsigned long)GetCurrentThreadId();
    return (ret);
}  

int init_locks(void)
{  
    int i;  


    lockarray = (HANDLE *)OPENSSL_malloc(CRYPTO_num_locks() * sizeof(HANDLE));
    if(NULL == lockarray)
    {
        return 1;
    }
    for (i=0; i<CRYPTO_num_locks(); i++)
    {
        lockarray[i] = CreateMutexA(NULL, 0, "");//lint !e409
    }

    CRYPTO_set_id_callback((unsigned long (*)())thread_id);  
    CRYPTO_set_locking_callback((void (*)(int, int, const char*, int))lock_callback);
    return 0;
}  

void kill_locks(void)  
{  
    int i;  

    CRYPTO_set_locking_callback(NULL);
    for (i=0; i<CRYPTO_num_locks(); i++)
    {
        CloseHandle(lockarray[i]);//lint !e409
    }

    OPENSSL_free(lockarray);
}  


void DisSleep(int Milliseconds)
{
    Sleep(Milliseconds);
}
// Openssl lock add by cwx298983 2016.07.29 End

#elif(OS_LINUX==OS_VER)

static pthread_mutex_t* lockarray;

unsigned int SmCreate( SM_T * const pSm )
{
    unsigned int err;

    if (NULL == pSm)
    {
        return OS_ERR_INVALID_PARAM;
    }

    err = pthread_mutex_init(&(pSm->mutex), NULL );
    if( err )
    {
        pthread_mutex_destroy (&(pSm->mutex));
        return OS_ERR_NOT_ENOUGH_MEM;
    }
    
    return OS_OK;
}

unsigned int SmDelete( SM_T *const pSm )
{
    unsigned int   err;

    if (NULL == pSm)
    {
        return OS_ERR_INVALID_PARAM;
    }

    err = 0;
    err |= pthread_mutex_destroy (&(pSm->mutex));

    if ( 0 != err )
    {
        err = OS_ERR_OBJ_ID_ERROR;
    }
    else
    {
        err = OS_OK;
    }

    return err;
}

unsigned int SmP( SM_T*const  pSm, unsigned int ulTimeOutInMillSec )
{
    unsigned int err;
    struct timespec       AbsTimeOut;
    err = 0;

    if (NULL == pSm)
    {
        return OS_ERR_INVALID_PARAM;
    }   

    
    err = pthread_mutex_lock (&(pSm->mutex));

    return err;
}

unsigned int SmV( SM_T* const pSm )
{
    unsigned int err;
    if (NULL == pSm)
    {
        return OS_ERR_INVALID_PARAM;
    }  

    err = pthread_mutex_unlock (&(pSm->mutex));


    return err;
}


unsigned int CreateTask
(  
    void(*pfnStartRoutine)(void*),
    void   *pArg, 
    unsigned int *pulTaskID 
)
{
    unsigned int     ulReturnCode;
    pthread_t        threadID;
    
    ulReturnCode = pthread_create( &threadID,
                                   NULL,
                                   ( _LINUX_TASK_ENTRY )pfnStartRoutine,
                                   ( void * )pArg 
                                  );

    if ( 0 != ulReturnCode ) 
    {
        *pulTaskID    = 0;
        return OS_ERR_NOT_ENOUGH_MEM;
    }
    /* LINUX reserve thread for expanding,
    and if LINUX expand itself, ulTaskId maybe only use one item of thread?*/
    *pulTaskID    = (unsigned int)threadID;
    return OS_OK;
}

unsigned int DeleteTask( unsigned int ulTaskID )
{
    if( pthread_self() == ulTaskID )
    {
        pthread_exit(0);
    }
    
    if( 0 != pthread_cancel( (pthread_t)ulTaskID ) )
    {
        return errno;
    }
    
    return OS_OK;
}

static void lock_callback(int mode, int type, char *file, int line)  
{
    (void)file;
    (void)line;
    if (mode & CRYPTO_LOCK) {
        pthread_mutex_lock(&(lockarray[type])); 
    }
    else
    {
        pthread_mutex_unlock(&(lockarray[type]));
    }  
}  

static unsigned long thread_id(void)  
{
    unsigned long ret;  

    ret = (unsigned long)pthread_self();

    return(ret);
}  

int init_locks(void)
{  
    int i;  
    lockarray = (pthread_mutex_t*)OPENSSL_malloc(CRYPTO_num_locks() * sizeof(pthread_mutex_t));
    if(NULL == lockarray)
    {
        return 1;
    }
    for (i=0; i<CRYPTO_num_locks(); i++)
    {
        pthread_mutex_init(&(lockarray[i]), NULL);
    }


    CRYPTO_set_id_callback((unsigned long (*)())thread_id);  
    CRYPTO_set_locking_callback((void (*)(int, int, const char*, int))lock_callback);
    return 0;
}  

void kill_locks(void)  
{  
    int i;  

    CRYPTO_set_locking_callback(NULL);
    for (i=0; i<CRYPTO_num_locks(); i++)
    {
        pthread_mutex_destroy(&(lockarray[i]));
    }

    OPENSSL_free(lockarray);
}  

void DisSleep(int Milliseconds)
{
    usleep(Milliseconds*1000);
}
// Openssl lock add by cwx298983 2016.07.29 End

#endif
#endif



