/** **************************************************************************
 * util.c
 * 
 * Copyright 2008 Bryan Ischo <bryan@ischo.com>
 * 
 * This file is part of libs3.
 * 
 * libs3 is free software: you can redistribute it and/or modify it under the
 * terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation, version 3 of the License.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of this library and its programs with the
 * OpenSSL library, and distribute linked combinations including the two.
 *
 * libs3 is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * version 3 along with libs3, in a file named COPYING.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 ************************************************************************** **/

#include <openssl/hmac.h>
#include <openssl/evp.h>
#include <ctype.h>
#include <string.h>
#include "util.h"
#include "discommon.h"
#include "log.h"
#include <pthread.h>
#include <sys/stat.h>
#include <stdlib.h>

#if 1
//lint -e26 -e31 -e63 -e64 -e78 -e101 -e119 -e129 -e144 -e156 -e438 -e505 -e515 -e516 -e522 -e529 -e530 -e533 -e534 -e546 -e551 -e578 -e601

#define rol(value, bits) (((value) << (bits)) | ((value) >> (32 - (bits))))

#define blk0L(i) (block->l[i] = (rol(block->l[i], 24) & 0xFF00FF00)     \
                  | (rol(block->l[i], 8) & 0x00FF00FF))

#define blk0B(i) (block->l[i])

#define blk(i) (block->l[i & 15] = rol(block->l[(i + 13) & 15] ^        \
                                       block->l[(i + 8) & 15] ^         \
                                       block->l[(i + 2) & 15] ^         \
                                       block->l[i & 15], 1))

#define R0_L(v, w, x, y, z, i)                                          \
    z += ((w & (x ^ y)) ^ y) + blk0L(i) + 0x5A827999 + rol(v, 5);       \
    w = rol(w, 30);
#define R0_B(v, w, x, y, z, i)                                          \
    z += ((w & (x ^ y)) ^ y) + blk0B(i) + 0x5A827999 + rol(v, 5);       \
    w = rol(w, 30);
#define R1(v, w, x, y, z, i)                                            \
    z += ((w & (x ^ y)) ^ y) + blk(i) + 0x5A827999 + rol(v, 5);         \
    w = rol(w, 30);
#define R2(v, w, x, y, z, i)                                            \
    z += (w ^ x ^ y) + blk(i) + 0x6ED9EBA1 + rol(v, 5);                 \
    w = rol(w, 30);
#define R3(v, w, x, y, z, i)                                            \
    z += (((w | x) & y) | (w & x)) + blk(i) + 0x8F1BBCDC + rol(v, 5);   \
    w = rol(w, 30);
#define R4(v, w, x, y, z, i)                                            \
    z += (w ^ x ^ y) + blk(i) + 0xCA62C1D6 + rol(v, 5);                 \
    w = rol(w, 30);

#define R0A_L(i) R0_L(a, b, c, d, e, i)
#define R0B_L(i) R0_L(b, c, d, e, a, i)
#define R0C_L(i) R0_L(c, d, e, a, b, i)
#define R0D_L(i) R0_L(d, e, a, b, c, i)
#define R0E_L(i) R0_L(e, a, b, c, d, i)

#define R0A_B(i) R0_B(a, b, c, d, e, i)
#define R0B_B(i) R0_B(b, c, d, e, a, i)
#define R0C_B(i) R0_B(c, d, e, a, b, i)
#define R0D_B(i) R0_B(d, e, a, b, c, i)
#define R0E_B(i) R0_B(e, a, b, c, d, i)

#define R1A(i) R1(a, b, c, d, e, i)
#define R1B(i) R1(b, c, d, e, a, i)
#define R1C(i) R1(c, d, e, a, b, i)
#define R1D(i) R1(d, e, a, b, c, i)
#define R1E(i) R1(e, a, b, c, d, i)

#define R2A(i) R2(a, b, c, d, e, i)
#define R2B(i) R2(b, c, d, e, a, i)
#define R2C(i) R2(c, d, e, a, b, i)
#define R2D(i) R2(d, e, a, b, c, i)
#define R2E(i) R2(e, a, b, c, d, i)

#define R3A(i) R3(a, b, c, d, e, i)
#define R3B(i) R3(b, c, d, e, a, i)
#define R3C(i) R3(c, d, e, a, b, i)
#define R3D(i) R3(d, e, a, b, c, i)
#define R3E(i) R3(e, a, b, c, d, i)

#define R4A(i) R4(a, b, c, d, e, i)
#define R4B(i) R4(b, c, d, e, a, i)
#define R4C(i) R4(c, d, e, a, b, i)
#define R4D(i) R4(d, e, a, b, c, i)
#define R4E(i) R4(e, a, b, c, d, i)



static void SHA1_transform(uint32_t state[5], const unsigned char buffer[64])
{
    uint32_t a, b, c, d, e;
    static uint32_t endianness_indicator = 0x1;

    typedef union {
        unsigned char c[64];
        uint32_t l[16];
    } u;

    unsigned char w[64];
    u *block = (u *) w;


    memcpy_S(block, 64, buffer, 64);


    a = state[0];
    b = state[1];
    c = state[2];
    d = state[3];
    e = state[4];

    
    if (((unsigned char *) &endianness_indicator)[0]) {
        R0A_L( 0);
        R0E_L( 1); R0D_L( 2); R0C_L( 3); R0B_L( 4); R0A_L( 5);
        R0E_L( 6); R0D_L( 7); R0C_L( 8); R0B_L( 9); R0A_L(10);
        R0E_L(11); R0D_L(12); R0C_L(13); R0B_L(14); R0A_L(15);
    }
    else {
        R0A_B( 0);
        R0E_B( 1); R0D_B( 2); R0C_B( 3); R0B_B( 4); R0A_B( 5);
        R0E_B( 6); R0D_B( 7); R0C_B( 8); R0B_B( 9); R0A_B(10);
        R0E_B(11); R0D_B(12); R0C_B(13); R0B_B(14); R0A_B(15);
    }
    R1E(16); R1D(17); R1C(18); R1B(19); R2A(20);
    R2E(21); R2D(22); R2C(23); R2B(24); R2A(25);
    R2E(26); R2D(27); R2C(28); R2B(29); R2A(30);
    R2E(31); R2D(32); R2C(33); R2B(34); R2A(35);
    R2E(36); R2D(37); R2C(38); R2B(39); R3A(40);
    R3E(41); R3D(42); R3C(43); R3B(44); R3A(45);
    R3E(46); R3D(47); R3C(48); R3B(49); R3A(50);
    R3E(51); R3D(52); R3C(53); R3B(54); R3A(55);
    R3E(56); R3D(57); R3C(58); R3B(59); R4A(60);
    R4E(61); R4D(62); R4C(63); R4B(64); R4A(65);
    R4E(66); R4D(67); R4C(68); R4B(69); R4A(70);
    R4E(71); R4D(72); R4C(73); R4B(74); R4A(75);
    R4E(76); R4D(77); R4C(78); R4B(79);

    state[0] += a;
    state[1] += b;
    state[2] += c;
    state[3] += d;
    state[4] += e;
}


typedef struct
{
    uint32_t state[5];
    uint32_t count[2];
    unsigned char buffer[64];
} SHA1Context;


static void SHA1_init(SHA1Context *context)
{
    context->state[0] = 0x67452301;
    context->state[1] = 0xEFCDAB89;
    context->state[2] = 0x98BADCFE;
    context->state[3] = 0x10325476;
    context->state[4] = 0xC3D2E1F0;
    context->count[0] = context->count[1] = 0;
}


static void SHA1_update(SHA1Context *context, const unsigned char *data,
                        unsigned int len)
{
    uint32_t i, j;

    j = (context->count[0] >> 3) & 63;

    if ((context->count[0] += len << 3) < (len << 3)) {
        context->count[1]++;
    }

    context->count[1] += (len >> 29);

    if ((j + len) > 63) {
        memcpy_S(&(context->buffer[j]), sizeof(context->buffer) - j, data, (i = 64 - j));
        SHA1_transform(context->state, context->buffer);
        for ( ; (i + 63) < len; i += 64) {
            SHA1_transform(context->state, &(data[i]));
        }
        j = 0;
    }
    else {
        i = 0;
    }

    memcpy_S(&(context->buffer[j]), sizeof(context->buffer) - j, &(data[i]), len - i);
}


static void SHA1_final(unsigned char digest[20], SHA1Context *context)
{
    uint32_t i;
    unsigned char finalcount[8] = {0};

    for (i = 0; i < 8; i++) {
        finalcount[i] = (unsigned char)
            ((context->count[(i >= 4 ? 0 : 1)] >>
              ((3 - (i & 3)) * 8)) & 255);
    }

    SHA1_update(context, (unsigned char *) "\200", 1);

    while ((context->count[0] & 504) != 448) {
        SHA1_update(context, (unsigned char *) "\0", 1);
    }

    SHA1_update(context, finalcount, 8);

    for (i = 0; i < 20; i++) {
        digest[i] = (unsigned char)
            ((context->state[i >> 2] >> ((3 - (i & 3)) * 8)) & 255);
    }

    memset_S(context->buffer, sizeof(context->buffer), 0, 64);
    memset_S(context->state, sizeof(context->state), 0, 20);
    memset_S(context->count, sizeof(context->count), 0, 8);
    //memset_s(&finalcount, sizeof(finalcount), 0, 8);
    memset_S(finalcount, sizeof(finalcount), 0, 8);

    SHA1_transform(context->state, context->buffer);
}


// HMAC-SHA-1:
//
// K - is key padded with zeros to 512 bits
// m - is message
// OPAD - 0x5c5c5c...
// IPAD - 0x363636...
//
// HMAC(K,m) = SHA1((K ^ OPAD) . SHA1((K ^ IPAD) . m))
void HMAC_SHA1(unsigned char hmac[20], const unsigned char *key, int key_len,
               const unsigned char *message, int message_len)
{
    unsigned char kopad[64] = {0};
    unsigned char kipad[64] = {0};
    int i;
    unsigned char digest[20];
    SHA1Context context;
    
    if (key_len > 64) {
        key_len = 64;
    }

    for (i = 0; i < key_len; i++) {
        kopad[i] = key[i] ^ 0x5c;
        kipad[i] = key[i] ^ 0x36;
    }

    for ( ; i < 64; i++) {
        kopad[i] = 0 ^ 0x5c;
        kipad[i] = 0 ^ 0x36;
    }


    
    SHA1_init(&context);
    SHA1_update(&context, kipad, 64);
    SHA1_update(&context, message, message_len);
    SHA1_final(digest, &context);

    SHA1_init(&context);
    SHA1_update(&context, kopad, 64);
    SHA1_update(&context, digest, 20);
    SHA1_final(hmac, &context);
    
}

#if 0
void HMAC_SHA256(unsigned char hmac[32], const unsigned char *key, int key_len,
               const unsigned char *message, int message_len)
{
    unsigned int tempLength = 0;
    HMAC_CTX ctx;
    unsigned int i;
    const EVP_MD* engine = EVP_sha256();

    // because the length of the string which output by the function is 32 bytes
    unsigned char* temp = (unsigned char*)malloc(32);
    if (NULL == temp)
    {
        return;
    }
    memset(temp, 0, 32);

    HMAC_CTX_init(&ctx);
    HMAC_Init_ex(&ctx, key, key_len, engine, NULL);
    HMAC_Update(&ctx, message, message_len);
    HMAC_Final(&ctx, temp, &tempLength);
    HMAC_CTX_cleanup(&ctx);

    memset(hmac, 0, 32);
    for(i=0; i<tempLength; i++)
    {
        hmac[i] = temp[i];
    }

    free(temp);
    temp = NULL;
}

void SHA256Hash(unsigned char sha[32], const unsigned char *message, int message_len)
{
    EVP_MD_CTX mdctx;
    const EVP_MD *md = NULL;
    unsigned char *temp = NULL;
    unsigned int tempLength = 0;
    unsigned int i = 0;

    OpenSSL_add_all_digests();

    md = EVP_get_digestbyname("sha256");
    if(NULL == md)
    {
        return ;
    }

    // because the length of the string which output by the function is 32 bytes
    temp = (unsigned char*)malloc(32);
    if (NULL == temp)
    {
        return;
    }
    memset(temp, 0, 32);


    EVP_MD_CTX_init(&mdctx);
    EVP_DigestInit_ex(&mdctx, md, NULL);
    EVP_DigestUpdate(&mdctx, message, message_len);
    EVP_DigestFinal_ex(&mdctx, temp, &tempLength);
    EVP_MD_CTX_cleanup(&mdctx);
    
    memset(sha, 0, 32);

    for(i = 0; i < tempLength; i++)
    {
        sha[i] = temp[i];
    }
    free(temp);
    temp = NULL;
}
#endif


void HMAC_SHA256(unsigned char hmac[32], const unsigned char *key, int key_len,
               const unsigned char *message, int message_len)
{
    unsigned int tempLength = 0;
    HMAC_CTX ctx;
    unsigned int i;
    const EVP_MD* engine = EVP_sha256();

    // because the length of the string which output by the function is 32 bytes
    /*
    unsigned char* temp = (unsigned char*)malloc(32);
    if (NULL == temp)
    {
        return;
    }
    memset(temp, 0, 32);
    */
    char temp[32] = {0};

    HMAC_CTX_init(&ctx);
    HMAC_Init_ex(&ctx, key, key_len, engine, NULL);
    HMAC_Update(&ctx, message, message_len);
    HMAC_Final(&ctx, temp, &tempLength);
    HMAC_CTX_cleanup(&ctx);

    memset(hmac, 0, 32);
    for(i=0; i<tempLength; i++)
    {
        hmac[i] = temp[i];
    }

    /*
    free(temp);
    temp = NULL;
    */
}

void SHA256Hash(unsigned char sha[32], const unsigned char *message, int message_len)
{  
  	// lock
  	if (pthread_mutex_lock(&mutex) != 0) {
  		printf("lock error!\n");
  	}
 
    EVP_MD_CTX mdctx;
    const EVP_MD *md = NULL;
    //unsigned char *temp = NULL;
    unsigned char temp[32] = {0};
    unsigned int tempLength = 0;
    unsigned int i = 0;

    OpenSSL_add_all_digests();

    md = EVP_get_digestbyname("sha256");
    if(NULL == md)
    {
        pthread_mutex_unlock(&mutex);
        return ;
    }

    // because the length of the string which output by the function is 32 bytes
    /*
    temp = (unsigned char*)malloc(32);
    if (NULL == temp)
    {
        return;
    }
    memset(temp, 0, 32);
    */


    EVP_MD_CTX_init(&mdctx);
    EVP_DigestInit_ex(&mdctx, md, NULL);
    EVP_DigestUpdate(&mdctx, message, message_len);
    EVP_DigestFinal_ex(&mdctx, temp, &tempLength);
    EVP_MD_CTX_cleanup(&mdctx);
    
    memset(sha, 0, 32);

    for(i = 0; i < tempLength; i++)
    {
        sha[i] = temp[i];
    }
    /*
    free(temp);
    temp = NULL;
    */
    // unlock
    pthread_mutex_unlock(&mutex);
}

#define rot(x,k) (((x) << (k)) | ((x) >> (32 - (k))))
//lint -e616
uint64_t hash(const unsigned char *k, int length)
{
    uint32_t a, b, c;
    static uint32_t endianness_indicator = 0x1;

    a = b = c = 0xdeadbeef + ((uint32_t) length);


    if (((unsigned char *) &endianness_indicator)[0]) {
        while (length > 12) {
            a += k[0];
            a += ((uint32_t) k[1]) << 8;
            a += ((uint32_t) k[2]) << 16;
            a += ((uint32_t) k[3]) << 24;
            b += k[4];
            b += ((uint32_t) k[5]) << 8;
            b += ((uint32_t) k[6]) << 16;
            b += ((uint32_t) k[7]) << 24;
            c += k[8];
            c += ((uint32_t) k[9]) << 8;
            c += ((uint32_t) k[10]) << 16;
            c += ((uint32_t) k[11]) << 24;
            a -= c; a ^= rot(c, 4);  c += b;
            b -= a; b ^= rot(a, 6);  a += c;
            c -= b; c ^= rot(b, 8);  b += a;
            a -= c; a ^= rot(c, 16);  c += b;
            b -= a; b ^= rot(a, 19);  a += c;
            c -= b; c ^= rot(b, 4);  b += a;
            length -= 12;
            k += 12;
        }
        
        switch(length) {
        case 12: c += ((uint32_t) k[11]) << 24;
        case 11: c += ((uint32_t) k[10]) << 16;
        case 10: c += ((uint32_t) k[9]) << 8;
        case 9 : c += k[8];
        case 8 : b += ((uint32_t) k[7]) << 24;
        case 7 : b += ((uint32_t) k[6]) << 16;
        case 6 : b += ((uint32_t) k[5]) << 8;
        case 5 : b += k[4];
        case 4 : a += ((uint32_t) k[3]) << 24;
        case 3 : a += ((uint32_t) k[2]) << 16;
        case 2 : a += ((uint32_t) k[1]) << 8;
        case 1 : a += k[0]; break;
        case 0 : goto end;
        }
    }
    else {
        while (length > 12) {
            a += ((uint32_t) k[0]) << 24;
            a += ((uint32_t) k[1]) << 16;
            a += ((uint32_t) k[2]) << 8;
            a += ((uint32_t) k[3]);
            b += ((uint32_t) k[4]) << 24;
            b += ((uint32_t) k[5]) << 16;
            b += ((uint32_t) k[6]) << 8;
            b += ((uint32_t) k[7]);
            c += ((uint32_t) k[8]) << 24;
            c += ((uint32_t) k[9]) << 16;
            c += ((uint32_t) k[10]) << 8;
            c += ((uint32_t) k[11]);
            a -= c; a ^= rot(c, 4);  c += b;
            b -= a; b ^= rot(a, 6);  a += c;
            c -= b; c ^= rot(b, 8);  b += a;
            a -= c; a ^= rot(c, 16);  c += b;
            b -= a; b ^= rot(a, 19);  a += c;
            c -= b; c ^= rot(b, 4);  b += a;
            length -= 12;
            k += 12;
        }

        switch(length) {
        case 12: c += k[11];
        case 11: c += ((uint32_t) k[10]) << 8;
        case 10: c += ((uint32_t) k[9]) << 16;
        case 9 : c += ((uint32_t) k[8]) << 24;
        case 8 : b += k[7];
        case 7 : b += ((uint32_t) k[6]) << 8;
        case 6 : b += ((uint32_t) k[5]) << 16;
        case 5 : b += ((uint32_t) k[4]) << 24;
        case 4 : a += k[3];
        case 3 : a += ((uint32_t) k[2]) << 8;
        case 2 : a += ((uint32_t) k[1]) << 16;
        case 1 : a += ((uint32_t) k[0]) << 24; break;
        case 0 : goto end;
        }
    }
    
    c ^= b; c -= rot(b, 14);
    a ^= c; a -= rot(c, 11);
    b ^= a; b -= rot(a, 25);
    c ^= b; c -= rot(b, 16);
    a ^= c; a -= rot(c, 4);
    b ^= a; b -= rot(a, 14);
    c ^= b; c -= rot(b, 24);

 end:
    return ((((uint64_t) c) << 32) | b);
}
//lint +e616
int is_blank(char c)
{
    return ((c == ' ') || (c == '\t'));
}

void uchar_to_hexes(unsigned char ucIn, unsigned char* szOut)
{
    static const unsigned char Number[] = 
    {
      '0', '1', '2', '3', '4', '5', '6', '7', 
      '8', '9', 'a', 'b', 'c', 'd', 'e', 'f'  
    };

    szOut[0] = Number[(ucIn & 0xF0) >> 4];
    szOut[1] = Number[(ucIn & 0x0F)];
}

void ustr_to_hexes(unsigned char* szIn,unsigned int inlen, unsigned char* szOut)
{
    unsigned int i;
    for (i = 0; i < inlen; ++i)
    {
        uchar_to_hexes(szIn[i], szOut + 2 * i);
    }
}
//lint +e26 +e31 +e63 +e64 +e78 +e101 +e119 +e129 +e144 +e156 +e438 +e505 +e516 +e515 +e522 +e529 +e530 +e533 +e534 +e546 +e551 +e578 +e601

#endif

void memset_S(void* dest, size_t destMax, int c, size_t count)
{
    size_t len = (destMax > count) ? count : destMax;
    memset(dest, c, len);
    return ;
}

void memcpy_S(void* dest, size_t destMax, const void* src, size_t count)
{
    size_t len = count > destMax ? destMax : count;
    memcpy(dest, src, len);
    return ;
}

unsigned int base64_encode(unsigned char *sourceStr, unsigned char* Base64Str, int sourcelen)  
{  
    long len;  
    long str_len;  
    unsigned char *res = Base64Str;  
    int i,j;  

    unsigned char *base64_table="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";  
  
	printfdbgfunc;

    str_len=sourcelen;  
    if(str_len % 3 == 0)  
        len=str_len/3*4;  
    else  
        len=(str_len/3+1)*4;  
  
  

    for(i=0,j=0;i<len-2;j+=3,i+=4)  
    {  
        res[i]=base64_table[sourceStr[j]>>2]; 
        res[i+1]=base64_table[(sourceStr[j]&0x3)<<4 | (sourceStr[j+1]>>4)]; 
        res[i+2]=base64_table[(sourceStr[j+1]&0xf)<<2 | (sourceStr[j+2]>>6)]; 
        res[i+3]=base64_table[sourceStr[j+2]&0x3f]; 
    }  
  
    switch(str_len % 3)  
    {  
        case 1:  
            res[i-2]='=';  
            res[i-1]='=';  
            break;  
        case 2:  
            res[i-1]='=';  
            break;  
    }  
  
	printfdbgoutfunc;

    return strlen(res);  
}  
  
unsigned int base64_decode(unsigned char *sourceStr, unsigned char *binStr)  
{  

    int table[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,62,0,0,0,63,52,53,54,55,56,57,58,59,60,61,0,0,0,0,0,0,0,0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,0,0,0,0,0,0,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,50,51};  
    long len;  
    unsigned int str_len;  
    unsigned char *res = binStr;  
    int i,j;  
  
	printfdbgfunc;

    len=strlen(sourceStr);  

    if(strstr(sourceStr,"=="))  
        str_len=len/4*3-2;  
    else if(strstr(sourceStr,"="))  
        str_len=len/4*3-1;  
    else  
        str_len=len/4*3;  

 
  

    for(i=0,j=0;i < len-2;j+=3,i+=4)  
    {  
        res[j]=((unsigned char)table[sourceStr[i]])<<2 | (((unsigned char)table[sourceStr[i+1]])>>4); 
        res[j+1]=(((unsigned char)table[sourceStr[i+1]])<<4) | (((unsigned char)table[sourceStr[i+2]])>>2); 
        res[j+2]=(((unsigned char)table[sourceStr[i+2]])<<6) | ((unsigned char)table[sourceStr[i+3]]); 
    }  
  
	printfdbgoutfunc;
    return str_len;
  
}  

/*
Calculate the file size
*/
int file_size(char* filename)
{
	struct stat statbuf;
	stat(filename, &statbuf);
	int size = statbuf.st_size;

	return size;
}

/**
* Create random UUID
*
* @param buf - buffer to be filled with the uuid string
*/
char *random_uuid(char buf[33])
{
	const char *c = "89ab";
	char *p = buf;
	int n;
	for (n = 0; n < 16; ++n)
	{
		int b = rand() % 255;
		switch (n)
		{
		case 6:
			sprintf(p, "4%x", b % 15);
			break;
		case 8:
			sprintf(p, "%c%x", c[rand() % strlen(c)], b % 15);
			break;
		default:
			sprintf(p, "%02x", b);
			break;
		}

		p += 2;
		//switch (n)
		//{
		//case 3:
		//case 5:
		//case 7:
		//case 9:
		//	*p++ = '-';
		//	break;
		//}
	}
	*p = 0;
	return buf;
}

